"""Neuromation auth client."""

from importlib.metadata import version

from .api import check_permissions, get_user, get_user_and_kind
from .client import (
    Action,
    AuthClient,
    ClientAccessSubTreeView,
    ClientSubTreeViewRoot,
    Permission,
    Role,
    User,
)
from .security import (
    JWT_CLUSTER_CLAIM,
    JWT_IDENTITY_CLAIM,
    JWT_KIND_CLAIM,
    IdentityError,
    Kind,
)

__all__ = [
    "Action",
    "AuthClient",
    "ClientAccessSubTreeView",
    "ClientSubTreeViewRoot",
    "JWT_IDENTITY_CLAIM",
    "JWT_KIND_CLAIM",
    "JWT_CLUSTER_CLAIM",
    "IdentityError",
    "Kind",
    "Permission",
    "Role",
    "User",
    "check_permissions",
    "get_user",
    "get_user_and_kind",
]
__version__ = version(__package__)



import dataclasses
import json
import logging
from collections.abc import Sequence
from typing import Union

from aiohttp import ClientError, web
from aiohttp_security import check_authorized
from aiohttp_security.api import AUTZ_KEY, IDENTITY_KEY

from .client import Permission
from .security import AuthPolicy, Kind

logger = logging.getLogger(__name__)


async def check_permissions(
    request: web.Request, permissions: Sequence[Union[Permission, Sequence[Permission]]]
) -> None:
    user_name = await check_authorized(request)
    auth_policy = request.config_dict.get(AUTZ_KEY)
    if not auth_policy:
        raise RuntimeError("Auth policy not configured")
    assert isinstance(auth_policy, AuthPolicy)

    try:
        missing = await auth_policy.get_missing_permissions(user_name, permissions)
    except ClientError as e:
        # re-wrap in order not to expose the client
        raise RuntimeError(e) from e

    if missing:
        payload = {"missing": [_permission_to_primitive(p) for p in missing]}
        raise web.HTTPForbidden(
            text=json.dumps(payload), content_type="application/json"
        )


@dataclasses.dataclass
class User:
    userid: str
    kind: Kind
    cluster: str | None  # non if kind != Kind.CLUSTER


async def get_user_and_kind(request: web.Request) -> tuple[str, Kind]:
    # Deprecated, user get_user() instead
    user = await get_user(request)
    return user.userid, user.kind


async def get_user(request: web.Request) -> User:
    identity_policy = request.config_dict.get(IDENTITY_KEY)
    if not identity_policy:
        raise RuntimeError("Identity policy not configured")
    auth_policy = request.config_dict.get(AUTZ_KEY)
    if not auth_policy:
        raise RuntimeError("Auth policy not configured")
    assert isinstance(auth_policy, AuthPolicy)
    identity = await identity_policy.identify(request)
    if identity is None:
        raise web.HTTPUnauthorized()
    userid = await auth_policy.authorized_userid(identity)
    if userid is None:
        raise web.HTTPUnauthorized()
    kind = auth_policy.get_kind(identity)
    cluster = auth_policy.get_cluster(identity) if kind is Kind.CLUSTER else None
    return User(userid, kind, cluster)


def _permission_to_primitive(perm: Permission) -> dict[str, str]:
    return {"uri": perm.uri, "action": perm.action}



import asyncio
from collections.abc import AsyncIterator, Mapping, Sequence
from contextlib import asynccontextmanager
from dataclasses import asdict, dataclass
from enum import Enum, unique
from typing import Any, Optional, Union

import aiohttp
from aiohttp.hdrs import AUTHORIZATION
from aiohttp.web import HTTPCreated, HTTPNoContent
from multidict import CIMultiDict, MultiDict
from yarl import URL

from .bearer_auth import BearerAuth


@dataclass(frozen=True)
class User:
    name: str
    # TODO (ajuszkowsi, March 2019) support "is_disabled" field


@unique
class Action(str, Enum):
    READ = "read"
    WRITE = "write"
    MANAGE = "manage"

    def __str__(self) -> str:
        return self.value

    def __repr__(self) -> str:
        return self.__str__().__repr__()


@dataclass(frozen=True)
class Permission:
    uri: str
    action: str

    def check_action_allowed(self, requested: str) -> bool:
        return check_action_allowed(self.action, requested)

    def can_list(self) -> bool:
        return check_action_allowed(self.action, "list")

    def can_read(self) -> bool:
        return check_action_allowed(self.action, "read")

    def can_write(self) -> bool:
        return check_action_allowed(self.action, "write")


@dataclass(frozen=True)
class Role:
    name: str
    permissions: Sequence[Permission]

    def to_role_permission(self, *, action: Action = Action.READ) -> Permission:
        return Permission(f"role://{self.name}", action)


@dataclass
class ClientAccessSubTreeView:
    action: str
    children: dict[str, "ClientAccessSubTreeView"]

    @classmethod
    def _from_json(cls, json_as_dict: dict[str, Any]) -> "ClientAccessSubTreeView":
        action = json_as_dict["action"]
        children = {
            name: ClientAccessSubTreeView._from_json(tree)
            for name, tree in json_as_dict["children"].items()
        }
        return ClientAccessSubTreeView(action, children)

    def check_action_allowed(self, requested: str) -> bool:
        return check_action_allowed(self.action, requested)

    def can_list(self) -> bool:
        return check_action_allowed(self.action, "list")

    def can_read(self) -> bool:
        return check_action_allowed(self.action, "read")

    def can_write(self) -> bool:
        return check_action_allowed(self.action, "write")


@dataclass
class ClientSubTreeViewRoot:
    scheme: str
    path: str
    sub_tree: ClientAccessSubTreeView

    def allows(self, perm: Permission) -> bool:
        perm_uri = URL(perm.uri)
        if perm_uri.scheme != self.scheme:
            return False
        perm_uri_no_scheme = (perm_uri.host or "") + perm_uri.path
        path_no_slash = self.path.lstrip("/").rstrip("/")
        if not perm_uri_no_scheme.startswith(path_no_slash):
            return False
        perm_uri_no_common_path = perm_uri_no_scheme[len(path_no_slash) :]
        if path_no_slash:
            if not perm_uri_no_common_path.startswith("/"):
                return False
            # [1:] to skip first "/":
            perm_uri_no_common_path = perm_uri_no_common_path[1:]
        parts = perm_uri_no_common_path.split("/")
        node = self.sub_tree
        for part in parts:
            if node.check_action_allowed(perm.action):
                return True
            if part not in node.children:
                return False
            node = node.children[part]
        return node.check_action_allowed(perm.action)

    @classmethod
    def _from_json(
        cls, scheme: str, json_as_dict: dict[str, Any]
    ) -> "ClientSubTreeViewRoot":
        subtree_path = json_as_dict["path"]
        sub_tree = ClientAccessSubTreeView._from_json(json_as_dict)
        return ClientSubTreeViewRoot(scheme, subtree_path, sub_tree)


class AuthClient:
    def __init__(
        self,
        url: Optional[URL],
        token: str,
        trace_configs: Optional[list[aiohttp.TraceConfig]] = None,
    ) -> None:
        if url is not None and not url:
            raise ValueError(
                "url argument should be http URL or None for secure-less configurations"
            )
        self._token = token
        headers = self._generate_headers(token)
        self._client = aiohttp.ClientSession(
            headers=headers, trace_configs=trace_configs
        )
        self._url = url

    async def __aenter__(self) -> "AuthClient":
        return self

    async def __aexit__(self, *args: Any) -> None:
        await self.close()

    def _generate_headers(self, token: Optional[str] = None) -> "CIMultiDict[str]":
        headers: CIMultiDict[str] = CIMultiDict()
        if token:
            headers[AUTHORIZATION] = BearerAuth(token).encode()
        return headers

    def _make_url(self, path: str) -> URL:
        assert self._url
        if path.startswith("/"):
            path = path[1:]
        return self._url / path

    async def close(self) -> None:
        await self._client.close()

    @asynccontextmanager
    async def _request(
        self,
        method: str,
        path: str,
        *,
        headers: Optional["CIMultiDict[str]"] = None,
        json: Any = None,
        params: Optional[Mapping[str, str]] = None,
        raise_for_status: bool = True,
    ) -> AsyncIterator[aiohttp.ClientResponse]:
        url = self._make_url(path)
        resp = await self._client.request(
            method, url, headers=headers, params=params, json=json
        )
        if raise_for_status:
            await _raise_for_status(resp)

        try:
            yield resp
        finally:
            resp.release()

    @property
    def is_anonymous_access_allowed(self) -> bool:
        return self._url is None

    async def ping(self) -> None:
        if self._url is None:
            return
        async with self._request("GET", "/api/v1/ping") as resp:
            txt = await resp.text()
            assert txt == "Pong"

    async def secured_ping(self, token: Optional[str] = None) -> None:
        if self._url is None:
            return
        path = "/api/v1/secured-ping"
        headers = self._generate_headers(token)
        async with self._request("GET", path, headers=headers) as resp:
            txt = await resp.text()
            assert txt == "Secured Pong"

    def _serialize_user(self, user: User) -> dict[str, Any]:
        return {"name": user.name}

    async def add_user(self, user: User, token: Optional[str] = None) -> None:
        path = "/api/v1/users"
        headers = self._generate_headers(token)
        payload = self._serialize_user(user)
        async with self._request("POST", path, headers=headers, json=payload):
            pass  # use context manager to release response earlier

    def _get_user_path(self, name: str) -> str:
        name = name.replace("/", ":")
        return f"/api/v1/users/{name}"

    async def get_user(self, name: str, token: Optional[str] = None) -> User:
        if self._url is None:
            return User(name="user")
        path = self._get_user_path(name)
        headers = self._generate_headers(token)
        async with self._request("GET", path, headers=headers) as resp:
            payload = await resp.json()
            return User(name=payload["name"])

    async def check_user_permissions(
        self,
        name: str,
        permissions: Sequence[Union[Permission, Sequence[Permission]]],
        token: Optional[str] = None,
    ) -> bool:
        if self._url is None:
            return True
        missing = await self.get_missing_permissions(name, permissions, token)
        return not missing

    async def get_missing_permissions(
        self,
        name: str,
        permissions: Sequence[Union[Permission, Sequence[Permission]]],
        token: Optional[str] = None,
    ) -> Sequence[Permission]:
        assert permissions, "No permissions passed"
        if self._url is None:
            return []
        path = self._get_user_path(name) + "/permissions/check"
        headers = self._generate_headers(token)
        payload: list[dict[str, Any]] = []
        has_alternatives = False
        for p in permissions:
            if isinstance(p, Permission):
                payload.append(asdict(p))
            else:
                has_alternatives = True
                for p2 in p:
                    payload.append(asdict(p2))
        async with self._request(
            "POST", path, headers=headers, json=payload, raise_for_status=False
        ) as resp:
            if resp.status not in (200, 403):
                await _raise_for_status(resp)
            data = await resp.json()
            if "missing" not in data:
                assert resp.status == 403, f"unexpected response {resp.status}: {data}"
                await _raise_for_status(resp)

            missing = [self._permission_from_primitive(p) for p in data["missing"]]
        if not missing or not has_alternatives:
            return missing
        required: set[Permission] = set()
        for p in permissions:
            if not isinstance(p, Permission):
                if all(p2 in missing for p2 in p):
                    required.update(p)
            else:
                required.add(p)
        return [p for p in missing if p in required]

    def _permission_from_primitive(self, perm: dict[str, str]) -> Permission:
        return Permission(uri=perm["uri"], action=perm["action"])

    async def get_permissions_tree(
        self, name: str, resource: str, depth: Optional[int] = None
    ) -> ClientSubTreeViewRoot:
        if self._url is None:
            return ClientSubTreeViewRoot(
                scheme=URL(resource).scheme,
                path="/default/user",
                sub_tree=ClientAccessSubTreeView(action="manage", children={}),
            )
        url = self._get_user_path(name) + "/permissions/tree"
        req_params: dict[str, Any] = {"uri": resource}
        if depth is not None:
            req_params["depth"] = depth
        async with self._request("GET", url, params=req_params) as resp:
            payload = await resp.json()
            tree = ClientSubTreeViewRoot._from_json(URL(resource).scheme, payload)
            return tree

    async def grant_user_permissions(
        self, name: str, permissions: Sequence[Permission], token: Optional[str] = None
    ) -> None:
        if self._url is None:
            return
        path = self._get_user_path(name) + "/permissions"
        headers = self._generate_headers(token)
        payload: list[dict[str, str]] = [asdict(p) for p in permissions]
        async with self._request("POST", path, headers=headers, json=payload) as resp:
            status = resp.status
            assert status == HTTPCreated.status_code, f"unexpected response: {status}"

    async def revoke_user_permissions(
        self, name: str, resources_uris: Sequence[str], token: Optional[str] = None
    ) -> None:
        if self._url is None:
            return
        path = self._get_user_path(name) + "/permissions"
        headers = self._generate_headers(token)
        params = MultiDict(("uri", uri) for uri in resources_uris)
        async with self._request(
            "DELETE", path, headers=headers, params=params
        ) as resp:
            status = resp.status
            assert status == HTTPNoContent.status_code, f"unexpected response: {status}"

    async def get_user_token(
        self,
        name: str,
        new_token_uri: Optional[str] = None,
        token: Optional[str] = None,
    ) -> str:
        if self._url is None:
            return ""
        path = self._get_user_path(name) + "/token"
        headers = self._generate_headers(token)
        if new_token_uri:
            data = {"uri": new_token_uri}
        else:
            data = {}
        async with self._request("POST", path, headers=headers, json=data) as resp:
            payload = await resp.json()
            return payload["access_token"]

    async def grant_role_permissions(
        self,
        role_name: str,
        permissions: Sequence[Permission],
        ignore_existing_role: bool = True,
    ) -> None:
        assert permissions
        if self._url is None:
            return
        try:
            await self.grant_user_permissions(role_name, permissions)
        except aiohttp.ClientResponseError as exc:
            if exc.status != 404:
                raise

            await self.add_user(User(name=role_name))
            await self.grant_user_permissions(role_name, permissions)

    async def set_role_permissions(
        self,
        role_name: str,
        permissions: Sequence[Permission],
        ignore_existing_role: bool = True,
    ) -> None:
        assert permissions
        if self._url is None:
            return
        try:
            await self._set_user_permissions(role_name, permissions)
        except aiohttp.ClientResponseError as exc:
            if exc.status != 404:
                raise

            await self.add_user(User(name=role_name))
            await self._set_user_permissions(role_name, permissions)

    async def _set_user_permissions(
        self, name: str, permissions: Sequence[Permission]
    ) -> None:
        path = self._get_user_path(name) + "/permissions"
        payload: list[dict[str, str]] = [asdict(p) for p in permissions]
        async with self._request("PUT", path, json=payload) as resp:
            status = resp.status
            assert (
                status == aiohttp.web.HTTPNoContent.status_code
            ), f"unexpected response: {status}"

    async def revoke_role_permissions(
        self, role_name: str, permissions: Sequence[Permission]
    ) -> None:
        assert permissions
        if self._url is None:
            return
        for perm in permissions:
            try:
                await self.revoke_user_permissions(role_name, [perm.uri])
            except aiohttp.ClientResponseError as exc:
                if exc.status == 400 and exc.message == "Operation has no effect":
                    pass
                else:
                    raise

    async def get_permissions(
        self, uname: str, *, expand_roles: bool = True
    ) -> Sequence[Permission]:
        if self._url is None:
            raise NotImplementedError("The method is not supported by Single-user mode")
        url = self._get_user_path(uname) + "/permissions"
        params = {}
        if not expand_roles:
            params["expand_roles"] = "false"
        async with self._request("GET", url, params=params) as resp:
            payload = await resp.json()
            return [Permission(item["uri"], item["action"]) for item in payload]

    async def add_role(self, uname: str) -> None:
        if self._url is None:
            return
        await self.add_user(User(name=uname))

    async def remove_role(self, uname: str) -> None:
        if self._url is None:
            return
        url = self._get_user_path(uname)
        try:
            async with self._request("DELETE", url):
                pass
        except aiohttp.ClientResponseError as exc:
            if exc.status != 404:
                raise

    async def delete_user(self, name: str, token: Optional[str] = None) -> None:
        if self._url is None:
            return
        path = self._get_user_path(name)
        headers = self._generate_headers(token)
        async with self._request("DELETE", path, headers=headers):
            pass  # use context manager to release response earlier


async def _raise_for_status(resp: aiohttp.ClientResponse) -> None:
    if 400 <= resp.status:
        details: str
        try:
            obj = await resp.json()
        except asyncio.CancelledError:
            raise
        except Exception:
            # ignore any error with reading message body
            details = resp.reason  # type: ignore
        else:
            try:
                details = obj["error"]
            except KeyError:
                details = str(obj)
        raise aiohttp.ClientResponseError(
            resp.request_info,
            resp.history,
            status=resp.status,
            message=details,
            headers=resp.headers,
        )


_action_order = {
    a: i for i, a in enumerate(("deny", "list", "read", "write", "manage"))
}


def _action_to_order(action: str) -> int:
    try:
        return _action_order[action]
    except KeyError:
        raise ValueError(f"invalid action: {action!r}") from None


def check_action_allowed(actual: str, requested: str) -> bool:
    return _action_to_order(actual) >= _action_to_order(requested)



from collections.abc import Sequence
from enum import Enum, StrEnum
from typing import Any, Optional, Union

from aiohttp.client_exceptions import ClientResponseError
from aiohttp.hdrs import AUTHORIZATION, SEC_WEBSOCKET_PROTOCOL
from aiohttp.helpers import BasicAuth
from aiohttp.web import Application, Request, StreamResponse
from aiohttp_security import AbstractAuthorizationPolicy, AbstractIdentityPolicy, setup
from jose import jwt
from jose.exceptions import JWTError

from .bearer_auth import BearerAuth
from .client import AuthClient, Permission, User

JWT_IDENTITY_CLAIM = "https://platform.neuromation.io/user"
JWT_IDENTITY_CLAIM_OPTIONS = ("identity", JWT_IDENTITY_CLAIM)
JWT_KIND_CLAIM = "https://platform.neuromation.io/kind"
# exists only for tokens of cluster services
JWT_CLUSTER_CLAIM = "https://platform.neuromation.io/cluster"

NEURO_AUTH_TOKEN_QUERY_PARAM = "neuro-auth-token"
WS_BEARER = "bearer.apolo.us-"


class AuthScheme(StrEnum):
    BASIC = "basic"
    BEARER = "bearer"


class Kind(StrEnum):
    CONTROL_PLANE = "control_plane"
    CLUSTER = "cluster"
    USER = "user"


class IdentityError(ValueError):
    """Identity error"""


class IdentityPolicy(AbstractIdentityPolicy):
    def __init__(
        self,
        auth_scheme: AuthScheme = AuthScheme.BEARER,
        default_identity: Optional[str] = None,
    ) -> None:
        self._auth_scheme = auth_scheme
        self._default_identity = default_identity

    async def identify(self, request: Request) -> Optional[str]:
        header_identity = request.headers.get(AUTHORIZATION)

        if header_identity is None:
            ws_identity = self._extract_ws_identity(request)
            query_identity = request.query.get(NEURO_AUTH_TOKEN_QUERY_PARAM)
            return ws_identity or query_identity or self._default_identity

        if self._auth_scheme == AuthScheme.BASIC:
            identity = BasicAuth.decode(header_identity).password
        else:
            identity = BearerAuth.decode(header_identity).token

        return identity

    async def remember(self, *args: Any, **kwargs: Any) -> None:  # pragma: no cover
        pass

    async def forget(
        self, request: Request, response: StreamResponse
    ) -> None:  # pragma: no cover
        pass

    def _extract_ws_identity(self, request: Request) -> str | None:
        ws_subprotocol = request.headers.get(SEC_WEBSOCKET_PROTOCOL)
        if ws_subprotocol is not None:
            for part in ws_subprotocol.strip().split(" "):
                if part.lower().startswith(WS_BEARER):
                    ws_identity = part[len(WS_BEARER) :]
                    return ws_identity
        return None


class AuthPolicy(AbstractAuthorizationPolicy):
    def __init__(self, auth_client: AuthClient) -> None:
        self._auth_client = auth_client

    def get_user_name_from_identity(self, identity: str | None) -> Optional[str]:
        if identity is None or self._auth_client.is_anonymous_access_allowed:
            return "user"

        try:
            claims = jwt.get_unverified_claims(identity)
            for identity_claim in JWT_IDENTITY_CLAIM_OPTIONS:
                if identity_claim in claims:
                    return claims[identity_claim]
            return None
        except JWTError:
            return None

    async def authorized_user(self, identity: str) -> Optional[User]:
        """Retrieve authorized user object (works same as authorized_userid)

        Return the user object identified by the identity
        or 'None' if no user exists related to the identity.
        """
        name = self.get_user_name_from_identity(identity)
        if not name:
            return None
        try:
            # NOTE: here we make a call to the auth service on behalf of the
            # actual user, not a service.
            return await self._auth_client.get_user(name, token=identity)
        except ClientResponseError:
            return None

    async def authorized_userid(self, identity: str) -> Optional[str]:
        user = await self.authorized_user(identity)
        if user:
            return user.name
        return None

    def get_kind(self, identity: str) -> Kind:
        try:
            claims = jwt.get_unverified_claims(identity)
            return Kind(claims.get(JWT_KIND_CLAIM, Kind.USER))
        except JWTError:
            return Kind.USER

    def get_cluster(self, identity: str) -> str:
        claims = jwt.get_unverified_claims(identity)
        kind = claims.get(JWT_KIND_CLAIM, Kind.USER)
        if kind != Kind.CLUSTER:
            raise IdentityError(
                f"identity has unexpected kind '{kind}', expected 'cluster'"
            )
        ret = claims.get(JWT_CLUSTER_CLAIM)
        if not ret:
            raise IdentityError(
                f"identity has kind 'cluster', but has no {JWT_CLUSTER_CLAIM} claim"
            )
        return ret

    async def permits(
        self,
        identity: str | None,
        permission: str | Enum,
        context: Any = None,
    ) -> bool:
        name = self.get_user_name_from_identity(identity)
        if not name:
            return False
        return await self._auth_client.check_user_permissions(name, context)

    async def get_missing_permissions(
        self,
        user_name: str,
        permissions: Sequence[Union[Permission, Sequence[Permission]]],
    ) -> Sequence[Permission]:
        return await self._auth_client.get_missing_permissions(user_name, permissions)


async def setup_security(
    app: Application,
    auth_client: AuthClient,
    auth_scheme: AuthScheme = AuthScheme.BEARER,
) -> None:  # pragma: no cover
    if auth_client.is_anonymous_access_allowed:
        identity_policy = IdentityPolicy(
            auth_scheme=auth_scheme, default_identity="user"
        )
    else:
        identity_policy = IdentityPolicy(auth_scheme=auth_scheme)
    auth_policy = AuthPolicy(auth_client=auth_client)
    setup(app, identity_policy, auth_policy)



from unittest.mock import Mock

import pytest
from aiohttp import BasicAuth
from aiohttp.test_utils import make_mocked_request
from jose import jwt

from neuro_auth_client.security import (
    JWT_CLUSTER_CLAIM,
    JWT_IDENTITY_CLAIM,
    JWT_KIND_CLAIM,
    AuthPolicy,
    AuthScheme,
    IdentityError,
    IdentityPolicy,
    Kind,
)


async def test_identity_bearer() -> None:
    req = make_mocked_request("GET", "/path", {"Authorization": "Bearer token1"})
    ip = IdentityPolicy()
    assert await ip.identify(req) == "token1"


async def test_identity_query() -> None:
    req = make_mocked_request("GET", "/path?neuro-auth-token=token2")
    ip = IdentityPolicy()
    assert await ip.identify(req) == "token2"


async def test_identity_default() -> None:
    req = make_mocked_request("GET", "/path")
    ip = IdentityPolicy(default_identity="default")
    assert await ip.identify(req) == "default"


async def test_identity_basic() -> None:
    auth = BasicAuth("user", "passwd")
    req = make_mocked_request("GET", "/path", {"Authorization": auth.encode()})
    ip = IdentityPolicy(AuthScheme.BASIC)
    assert await ip.identify(req) == "passwd"


async def test_identity_ws() -> None:
    req = make_mocked_request(
        "GET", "/path", {"Sec-WebSocket-Protocol": "bearer.apolo.us-token4"}
    )
    ip = IdentityPolicy()
    assert await ip.identify(req) == "token4"


async def test_identity_ws_multiple_subprotocols() -> None:
    req = make_mocked_request(
        "GET", "/path", {"Sec-WebSocket-Protocol": "chat bearer.apolo.us-token4 other"}
    )
    ip = IdentityPolicy()
    assert await ip.identify(req) == "token4"


@pytest.mark.parametrize(
    "path,headers,token",
    [
        (  # all
            "/path?neuro-auth-token=token1",
            {
                "Authorization": "Bearer token2",
                "Sec-WebSocket-Protocol": "bearer.apolo.us-token3",
            },
            "token2",
        ),
        (  # header + query
            "/path?neuro-auth-token=token1",
            {
                "Authorization": "Bearer token2",
            },
            "token2",
        ),
        (  # header + ws
            "/path",
            {
                "Authorization": "Bearer token2",
                "Sec-WebSocket-Protocol": "bearer.apolo.us-token3",
            },
            "token2",
        ),
        (  # query + ws
            "/path?neuro-auth-token=token1",
            {
                "Sec-WebSocket-Protocol": "bearer.apolo.us-token3",
            },
            "token3",
        ),
        (  # default
            "/path",
            {},
            "default",
        ),
    ],
)
async def test_identity_precedence(
    path: str, headers: dict[str, str], token: str
) -> None:
    req = make_mocked_request("GET", path, headers)
    ip = IdentityPolicy(default_identity="default")
    assert await ip.identify(req) == token


def test_get_kind() -> None:
    identity = jwt.encode(
        {JWT_IDENTITY_CLAIM: "test", JWT_KIND_CLAIM: Kind.CLUSTER}, "secret"
    )
    auth_policy = AuthPolicy(Mock())
    assert auth_policy.get_kind(identity) == Kind.CLUSTER


def test_get_kind_default() -> None:
    identity = jwt.encode({JWT_IDENTITY_CLAIM: "test"}, "secret")
    auth_policy = AuthPolicy(Mock())
    assert auth_policy.get_kind(identity) == Kind.USER


def test_get_cluster() -> None:
    identity = jwt.encode(
        {
            JWT_IDENTITY_CLAIM: "test",
            JWT_KIND_CLAIM: "cluster",
            JWT_CLUSTER_CLAIM: "cl1",
        },
        "secret",
    )
    auth_policy = AuthPolicy(Mock())
    assert auth_policy.get_cluster(identity) == "cl1"


def test_get_cluster_invalid_kind() -> None:
    identity = jwt.encode(
        {JWT_IDENTITY_CLAIM: "test", JWT_KIND_CLAIM: "user", JWT_CLUSTER_CLAIM: "cl1"},
        "secret",
    )
    auth_policy = AuthPolicy(Mock())
    with pytest.raises(IdentityError, match="^identity has unexpected kind"):
        auth_policy.get_cluster(identity)


def test_get_cluster_invalid_claim() -> None:
    identity = jwt.encode(
        {JWT_IDENTITY_CLAIM: "test", JWT_KIND_CLAIM: "cluster", JWT_CLUSTER_CLAIM: ""},
        "secret",
    )
    auth_policy = AuthPolicy(Mock())
    with pytest.raises(IdentityError, match="^identity has kind 'cluster', but"):
        auth_policy.get_cluster(identity)
